#!/bin/bash


# version 0.5. 20230213. ken woo. copyleft.

# from 0.2:
# 1. bug fixed [sed -n '/Date/,/files/p'] to [sed -n '/Date/,/\lfile/p'].
# 2. found the $IFS bug but not yet resolved.
# from 0.2a:
# 1. the altering $IFS was changed into read -r for grep processing [grep $2 -- "$ctx"].
# from 0.2b:
# 1. added time elapsed.
# 2. added a 'cp' identifier for use in $2 if want to copy to $3 directly without context match just only filename match.
# 3. added to generate a path file besides of a target file, hence this target file could be easily found manually.
# from 0.3: only add some comments.
# from 0.3r:
# 1. fixed [sed -n '/Date/,/\lfile/p'] even it affects nothing(false negative).
# 2a. changed the sed command into correct one since it sometimes failed match, [tmp_str=$( echo "$tt" | sed 's,'"$TMP_DIR"',,' )].
# 2b. it only affects the path in "path file" which would lose some intermediate path-string. the correct one uses bash internal cmd.
# 3. added "$2"=="cq" for only copying path files without target files. (since this way is simpler than outputing via std-out).
# from 0.4:
# 1. fixed [ while [ -e "${3}${k}" ] ] not concerning about path file.
# 2. added "$2"=="cr" for adding sha1sum checksum file *.cksm along with path file. to separate is easy access, e.g., "cat *.cksm".

# purposes:
# 1. dig out all the "$1" files(only show results of file path/name) under current working dir/subdir and
#     including which packed within archive files even within archive of archive of archive of...
#
# 2. the same as 1., and also search contents of the matched files by pattern "$2", which is also in grep style pattern and
#     including options, e.g., $2==".*void +main\(.*\) -i -n -B5 -A5". (note, the options in trailing is must).
#
# 3. the same as 2., and copy the matched files to the assigned absolute-directory by "$3".
#
# 4. the same as 3., but directly copies without specifying context pattern in $2, instead, set $2 as 'cp' for this purpose.
#     note, you can try ".*" in $2 in comparison with "cp". another is 'cq' for copying path file only. yet another 'cr'.
#
# 5a. another path file(*.path) is coming up with the copied target file which can be easily found manually via this path file.
# 5b. use $2=='cr' to not only 'cq' but also add checksum in *.cksm file.

# required:
# 1. apt install p7zip-full p7zip-rar
# 2. in the /tmp/ dir for extracting files for temporary use, auto deleted after done.

# known limitation:
# 1. the search pattern is in grep format.
# 2. search twice if the soft-link and its real-path are both in the searching-paths.
# 3. multi-volume archive will cause 7z to exit. so rename existing ones are needed before running this script into completely.
# 4. it is used by "7z -px" to skip password protected archives.
# 5. it is used by "7z -aou" to auto rename repetitive extracted files.
# 6. token 'null' is used, so it is not for pattern search. so is the 'cp', 'cq', and 'cr' in $2.
# 7. file descriptor 3 is used for identification purpose, so do not use it before running this script.
# 8. except parameters $1, $2, $3 are for user, the other params are only for internal use:
#     "$4" passing a specific archive file for extracting and searching.
#     "$5" script path.
#     "$6" script name.
#     "$7" progressive path for inheriting.
# 9. $2, the content pattern, must have the options in trailing. that is like as "the_pattern -opt1 -opt2".
# 10. the target path for copying can not be in the searching path or it might be searched and cause loop.
# 11a. $2 has 4 identifiers 'null' and 'c[pqr]' can not be used, however, still could be the pattern by e.g., "nul[l]\{1\}".
# 11b. note, weird: [grep nul[l]\{1\} -n --color a.txt] is failed and [this_script ".*a\.txt.*" "nul[l]\{1\} -n --color"] is success.
# 11c. $1 needs full qualified name, that is, if "a.txt" is exactly the file to search, it still needs to be like as ".*a\.txt.*"
# 11d. the 11c might be the reason of prefixed path.
# 12. known fault: the filenames to inspect if containing 1) comma "," 2) leading hyphen "-" etc., either made sed fault or grep fault.

# example1: [this_script ".*\.pdf$"]
#     find out all the "pdf" files where locate including sub-dirs and within archive files.
#
# example2: detach the task from shell for free run.
# [sudo nohup this_sh ".*\.pdf$\|.*\.doc$\|.*\.chm$\|.*\.djvu$\|.*\.pptx\?$\|.*\.pps$\|.*\.xlsx\?$\|.*\.mht$" &> ~/Output.txt & disown]
#
# example3: [this_script ".*\.c$" ".*printf.* -n -B3 -A5"]
#     find out all the "*.c" files which have pattern '.*printf.*' in it. and
#     also dump out these line-number and the before 3 lines and the after 5 lines.
#
# example4: [this_script ".*\.c$" ".*printf.* -n -B3 -A5" /home/user/Desktop/target]
#     the same as example3 and copy matched files to folder /home/user/Desktop/target/.
#
# example5: [this_script .*\.txt$ "[^[:blank:]]\+Group[[:blank:]]\{1\} -n -i -A5 -B4 --color" /home/user/Desktop/temp/]
#
# example6: [this_script .*\.txt$ cp /home/user/Desktop/temp/]
#     all the found *.txt files with .path files will directly copy to /home/user/Desktop/temp/.
#
# example7: "the_copied_target.file" has a companion file "the_copied_target.file.path" unconditionally. so, rm *.path if not required.
#
# example8: [cat outputfile | sed '/^-\+>\+\|^[\(Open \)\(ERROR\)\(WARNINGS\)]\|^Is not archive/d;/^$/d;s/\(\/.*\/\)\([^\/]\+$\)/\2/']
#     extract only filenames from result output. note, 7z-error/-open-error files may just not supported, so need to treat by other ways.
#
# example9: [this_script ".*\.\(zip\|rar\|ace\|arj\|t\?gz\|tar\|z\|7z\|bz2\|lzh\|txt\|ex[e_]\{1\}\|bat\|msi\|sys\|dll\|ocx\|bin\|pdf\|djvu\|html\?\|mht\|chm\|css\|js\|ps\|iso\|nrg\|gho\|cab\|avi\|rm\(vb\)\?\|jpe\?g\|bmp\|ico\|gif\|mp[34]\{1\}\|mpe\?g\|png\|doc\|ini\|hlp\|inf\|ttf\|pptx\?\|pps\|xlsx\?\|reg\|dat\|db\|bak\|log\|asm\|inc\|c\(pp\|xx\)\?\|h\(pp\)\?\|lib\)$" &> output.txt]


SERF_VER="0.5"
lsof -a -p $$ -d 3 2>/dev/null | grep -i -q "\ 3w\ \|\ 3u\ "    # use file descriptor 3 for recognizing root

if [ $? -eq 0 ]; then                                   # if not the root task

    if [ $# -eq 7 ]; then                               # if intends for fork-task
        inipath=$( pwd )
        arcname="$( basename "$4" ).XXXXXXXX"
        arcpath=$( mktemp -t -d "$arcname" )
        SCRIPT_DIR="$5"
        SCRIPT_NAME="$6"
        INI_DIR="$7"
        TMP_DIR="$arcpath"
        7z -px -aou -o"$arcpath" x "$4" 2>/dev/null 1>/dev/null
        cd "$arcpath"
    else                                                # something wrong
        exit 1
    fi

else                                                    # the root task, generates fd3

    # trap prevents from aborted by user and $IFS not yet recovered
    IFS_OLD="$IFS"
    function for_trap_exit() {
        echo -e '\n\nuser abort\n';
        [[ "$IFS_OLD" != "$IFS" ]] && IFS="$IFS_OLD" && echo clean up;
        # be careful not to clean-up /tmp/ here or get into catastrophe since I met.
        exec 3>&-;
        exit 1;
    }
    trap for_trap_exit SIGINT SIGKILL SIGSEGV

    # wrong arguments
    if [ $# -lt 1 ] || [ $# -gt 3 ]; then
        echo; echo "usage: ${0} \"filename pattern for search\" [\"context pattern in file\" [\"path to extract if match\"] ]";
        exit 1
    fi

    # no 7z bin file
    7z | grep -i "copyright"
    if [ $? -ne 0 ]; then
        echo; echo "utility 7z is needed. please try \"apt install p7zip-full p7zip-rar\" first.";
	exit 1
    fi

    # set global vars
    SCRIPT_DIR=$( cd -- "$( dirname -- "${BASH_SOURCE[0]}" )" &> /dev/null && pwd )
    SCRIPT_NAME=$( basename -- "${BASH_SOURCE[0]}" )
    TMP_DIR="presently working directory"
    INI_DIR=""
    inipath=$( pwd )
    arcpath=$( pwd )

    # rearrange arguments
    z1="$1"
    if [ $# -eq 1 ]; then
        z2='null'
        z3='null'
    elif [ $# -eq 2 ]; then
        if [ "$2" = 'null' ]; then
            echo '"null" not applicable.'
            exit 1
        elif [ "$2" = cp ] || [ "$2" = cq ] || [ "$2" = cr ]; then
            echo "the absolute path is needed."
            exit 1
        fi
        z2="$2"
        z3='null'
    else
        if [ "$2" = 'null' ] || [ "$3" = 'null' ]; then
            echo '"null" not applicable.'
            exit 1
        fi
        z2="$2"
        z3="$3"
        echo "$z3" | grep -q "^/.*"
        if [ $? -eq 0 ] && [ -d "$z3" ]; then

            # try to padding "/"
            echo "$z3" | grep -q ".*/$"
            [[ $? -ne 0 ]] && z3="${z3}/"

        else
            echo; echo "the absolute path is needed."; echo;
            exit 1
        fi
    fi

    set --
    set "$z1" "$z2" "$z3"
    exec 3>&1

    echo -e "\nscript version: $SERF_VER\n"
    echo "cmd: $SCRIPT_DIR/$SCRIPT_NAME '$1' '$2' '$3'"
    echo "pwd: $inipath"

    echo; echo $( date ); echo;
    elapse_time_b=$SECONDS

fi


if [ "$2" != $'null' ]; then                            # needs for context search or direct-copy
    find -L ~+ -type f -regextype grep -iregex "$1" | while read -r ctx;
    do

        # note, by the grep, the $2 not quoted for options injection
        [[ "$2" = 'cp' ]] || [[ "$2" = 'cq' ]] || [[ "$2" = 'cr' ]] || grep $2 -- "$ctx"

        if [ $? -eq 0 ]; then
            echo "$ctx    [[[Found Here]]]"; echo -e "----------------\n";
            if [ "$3" != $'null' ]; then
                i=0;
                j=$( basename -- "$ctx" )
                k=$j
                while [ -e "${3}${k}" ] || [ -e "${3}${k}.path" ]; do
                    (( i=$i+1 ));
		    k="$j($i)"
                done
                [[ "$2" = 'cp' ]] && cp "$ctx" "${3}${k}"   # copy the matched file to the target folder

                # generate the path of the copied file
                # note the current file is in /tmp/arc-name/path-in-arc, so we need path-in-arc string
                # however it might be in the $PWD instead, so use sed to cover both conditions
                #tmp_str=$( echo "$ctx" | sed 's,'"$TMP_DIR"',,' )    # remove the "/tmp/arc-name"
                # the previous sed command sometimes causes failed match since some special chars still not escape,
                # so changed to the following line
                tmp_str=${ctx##"$TMP_DIR"}

		echo "${INI_DIR}${tmp_str}" > "${3}${k}.path"
                [[ "$2" = 'cr' ]] && echo $( sha1sum "$ctx" ) | sed 's,\ .*,\ '"${INI_DIR}${tmp_str}"',' > "${3}${k}.cksm"
            fi
        fi

    done
else                                                    # only for file name search
    find -L ~+ -type f -regextype grep -iregex "$1" | sed 's/\(.*\)/\"\1\"/' | xargs -n1 # list all the existing pdfs
fi


t=$( find -L ~+ -type f -regextype sed -iregex '.*\.\(zip\|rar\|ace\|arj\|t\?gz\|tar\|z\|7z\|xz\|bz2\|lzh\|ex[e_]\{1\}\|iso\)' | sed 's/\(.*\)/\"\1\"/' | xargs -n1 )

IFS_OLD="$IFS"
IFS=$'\n'
for tt in $t; do

    echo -e "----->>>>>  $tt\n"

    # the following line is incorrect but is ok since it through to the end(where meant to be) and "\lfile" meant to be not found.
    # the new replacement command however weird, when use 1 instead of 0 would go wrong.
    #ttt=$( 7z -px l "$tt" | sed -n '/Date/,/\lfile/p' )
    ttt=$( 7z -px l "$tt" | tac | sed '0,/fi/d' | tac | sed '0,/Dat/d' )
    v=$( echo "$ttt" | grep -m1 -q -i "\.zip$\|\.rar$\|\.ace$\|\.arj$\|\.t\?gz$\|\.tar$\|\.z$\|\.7z$\|\.xz$\|\.bz2$\|\.lzh$\|\.ex[e_]\{1\}$\|\.iso$" )
    if [ $? -eq 0 ]; then                               # need to extract further for advanced search since another package inside
        IFS="$IFS_OLD"                                  # branch out then need recovery

        # note the current file is in /tmp/arc-name/path-in-arc, so we need path-in-arc string
        # however it might be in the $PWD instead, so use sed to cover both conditions
        #tmp_str=$( echo "$tt" | sed 's,'"$TMP_DIR"',,' )
        # the previous sed command sometimes causes failed match since some special chars still not escape,
        # so changed to the following line
        tmp_str=${tt##"$TMP_DIR"}

        "$SCRIPT_DIR"/"$SCRIPT_NAME" "$1" "$2" "$3" "$tt" "$SCRIPT_DIR" "$SCRIPT_NAME" "${INI_DIR}${tmp_str}"

        IFS=$'\n'                                       # branch in then need recovery
    elif [ "$2" != $'null' ]; then                      # need to extract further if at least one match for context search
        echo "$ttt" | grep -m1 -q -i "$1"
        if [ $? -eq 0 ]; then
            IFS="$IFS_OLD"

            # note the current file is in /tmp/arc-name/path-in-arc, so we need path-in-arc string
            # however it might be in the $PWD instead, so use sed to cover both conditions
            #tmp_str=$( echo "$tt" | sed 's,'"$TMP_DIR"',,' )
            # the previous sed command sometimes causes failed match since some special chars still not escape,
            # so changed to the following line
            tmp_str=${tt##"$TMP_DIR"}

            "$SCRIPT_DIR"/"$SCRIPT_NAME" "$1" "$2" "$3" "$tt" "$SCRIPT_DIR" "$SCRIPT_NAME" "${INI_DIR}${tmp_str}"

            IFS=$'\n'
	fi
    else
        echo "$ttt" | grep -i "$1"
    fi

done
IFS="$IFS_OLD"

if [ "$inipath" != "$arcpath" ]; then                   # if not the root
    cd "$inipath"
    rm -rf "$arcpath"
else
    exec 3>&-                                           # done the search and release fd3
    echo; echo $( date ); echo;
    (( time_elapsed=$SECONDS-$elapse_time_b ));
    echo -e "\nit took $(( $time_elapsed / 60 )) minute(s) $(( $time_elapsed % 60 )) seconds\n";
fi


# end of sh
